"""Wrappers around standard library multiprocessing executors."""
import concurrent.futures
import multiprocessing
import threading


class _Executor(object):

    def submit(self, fn, *args, **kwargs):
        self.semaphore.acquire()
        future = super().submit(fn, *args, **kwargs)
        future.add_done_callback(self._release)

        return future

    def _release(self, future):
        self.semaphore.release()


class ProcessPoolExecutor(_Executor, concurrent.futures.ProcessPoolExecutor):
    """Extends `ProcessPoolExecutor` by limiting simultaneous work items."""

    def __init__(self, max_items, **kwargs):
        """Initialize a process pool.

        :param max_items: maximum number of simultaneous work items.
            Calls to `.submit` will block if there are too many unprocessed
            items.
        :param kwargs: key-word arguments to `ProcessPoolExecutor`.
        """
        super().__init__(**kwargs)
        self.semaphore = multiprocessing.BoundedSemaphore(max_items)


class ThreadPoolExecutor(_Executor, concurrent.futures.ThreadPoolExecutor):
    """Extends `ThreadPoolExecutor` by limiting simultaneous work items."""

    def __init__(self, max_items, **kwargs):
        """Initialize a thread pool.

        :param max_items: maximum number of simultaneous work items.
            Calls to `.submit` will block if there are too many unprocessed
            items.
        :param kwargs: key-word arguments to `ThreadPoolExecutor`.
        """
        super().__init__(**kwargs)
        self.semaphore = threading.BoundedSemaphore(max_items)



"""Commonly used data structures and functions."""
import collections
import logging
import os

import pysam


# provide read only access to key region attrs
_Region = collections.namedtuple('Region', 'ref_name start end')


class Region(_Region):
    """Represents a genomic region."""

    @property
    def name(self):
        """Samtools-style region string, zero-base end exclusive."""
        return self.__str__()

    def __str__(self):
        """Return string representation of region."""
        # This will be zero-based, end exclusive
        start = 0 if self.start is None else self.start
        end = '' if self.end is None else self.end
        return '{}:{}-{}'.format(self.ref_name, start, end)

    @property
    def size(self):
        """Return size of region."""
        return self.end - self.start

    @classmethod
    def from_string(cls, region):
        """Parse region string into `Region` objects.

        :param region: region str

        >>> Region.from_string('Ecoli') == Region(
        ...     ref_name='Ecoli', start=None, end=None)
        True
        >>> Region.from_string('Ecoli:1000-2000') == Region(
        ...     ref_name='Ecoli', start=1000, end=2000)
        True
        >>> Region.from_string('Ecoli:1000') == Region(
        ...     ref_name='Ecoli', start=1000, end=None)
        True
        >>> Region.from_string('Ecoli:-1000') == Region(
        ...     ref_name='Ecoli', start=0, end=1000)
        True
        >>> Region.from_string('Ecoli:500-') == Region(
        ...     ref_name='Ecoli', start=500, end=None)
        True
        >>> Region.from_string('A:B:c:500-') == Region(
        ...     ref_name='A:B:c', start=500, end=None)
        True
        """
        if ':' not in region:
            ref_name, start, end = region, None, None
        else:
            start, end = None, None
            ref_name, bounds = region.rsplit(':', 1)
            if bounds[0] == '-':
                start = 0
                end = int(bounds.replace('-', ''))
            elif '-' not in bounds:
                start = int(bounds)
                end = None
            elif bounds[-1] == '-':
                start = int(bounds[:-1])
                end = None
            else:
                start, end = [int(b) for b in bounds.split('-')]
        return cls(ref_name, start, end)

    def split(region, size, overlap=0, fixed_size=True):
        """Split region into sub-regions of a given length.

        :param size: size of sub-regions.
        :param overlap: overlap between ends of sub-regions.
        :param fixed_size: ensure all sub-regions are equal in size. If `False`
            then the final chunk will be created as the smallest size to
            conform with `overlap`.

        :returns: a list of sub-regions.

        """
        regions = list()
        for start in range(region.start, region.end, size - overlap):
            end = min(start + size, region.end)
            regions.append(Region(region.ref_name, start, end))
        if len(regions) > 1:
            if fixed_size and regions[-1].size < size:
                del regions[-1]
                end = region.end
                start = end - size
                if start > regions[-1].start:
                    regions.append(Region(region.ref_name, start, end))
        return regions

    def overlaps(self, other):
        """Determine if a region overlaps another.

        :param other: a second Region to test overlap.

        :returns: True if regions overlap.

        """
        if self.ref_name != other.ref_name:
            return False

        def _limits(x):
            x0 = x.start if x.start is not None else -1
            x1 = x.end if x.end is not None else float('inf')
            return x0, x1

        a0, a1 = _limits(self)
        b0, b1 = _limits(other)
        return (
            (a0 < b1 and a1 > b0) or
            (b0 < a1 and b1 > a0))


def get_regions(bam, region_strs=None):
    """Create `Region` objects from a bam and region strings.

    :param bam: `.bam` file.
    :param region_strs: iterable of str in zero-based (samtools-like)
        region format e.g. ref:start-end or filepath containing a
        region string per line.

    :returns: list of `Region` objects.
    """
    with pysam.AlignmentFile(bam) as bam_fh:
        ref_lengths = dict(zip(bam_fh.references, bam_fh.lengths))
    if region_strs is not None:
        if os.path.isfile(region_strs[0]):
            with open(region_strs[0]) as fh:
                region_strs = [line.strip() for line in fh.readlines()]

        regions = []
        for r in (Region.from_string(x) for x in region_strs):
            start = r.start if r.start is not None else 0
            end = r.end if r.end is not None else ref_lengths[r.ref_name]
            regions.append(Region(r.ref_name, start, end))
    else:
        regions = [
            Region(ref_name, 0, end)
            for ref_name, end in ref_lengths.items()]

    return regions


def get_named_logger(name):
    """Create a logger with a name."""
    logger = logging.getLogger('{}.{}'.format(__package__, name))
    logger.name = name
    return logger


comp = {
    'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 'X': 'X', 'N': 'N',
    'a': 't', 't': 'a', 'c': 'g', 'g': 'c', 'x': 'x', 'n': 'n',
    # '-': '-'
}
comp_trans = str.maketrans(''.join(comp.keys()), ''.join(comp.values()))


def reverse_complement(seq):
    """Reverse complement sequence.

    :param: input sequence string.

    :returns: reverse-complemented string.

    """
    return seq.translate(comp_trans)[::-1]



import os
import re
import sys
from glob import glob
import importlib.util
import shutil
from setuptools import setup, find_packages, Extension
from setuptools import Distribution, Command
from setuptools.command.install import install
from setuptools.command.build_ext import build_ext
import subprocess

__pkg_name__ = 'fast5mod'
__author__ = 'ont-research'
__description__ = 'Extraction of modified base data from Guppy Fast5 output'

# Use readme as long description and say its github-flavour markdown
from os import path
this_directory = path.abspath(path.dirname(__file__))
kwargs = {'encoding':'utf-8'} if sys.version_info.major == 3 else {}
with open(path.join(this_directory, 'README.md'), **kwargs) as f:
    __long_description__ = f.read()
__long_description_content_type__ = 'text/markdown'

__path__ = os.path.dirname(__file__)
__pkg_path__ = os.path.join(os.path.join(__path__, __pkg_name__))

# find version
verstrline = open(os.path.join(__pkg_name__, '__init__.py'), 'r').read()
vsre = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(vsre, verstrline, re.M)
if mo:
    __version__ = mo.group(1)
else:
    raise RuntimeError('Unable to find version string in "{}/__init__.py".'.format(__pkg_name__))

# create requirements from requirements.txt
dir_path = os.path.dirname(__file__)
install_requires = []
with open(os.path.join(dir_path, 'requirements.txt')) as fh:
    reqs = (
        r.split('#')[0].strip()
        for r in fh.read().splitlines() if not r.strip().startswith('#')
    )
    for req in reqs:
        if req.startswith('git+https'):
            req = req.split('/')[-1].split('@')[0]
        install_requires.append(req)


if __name__ == '__main__':

    pymajor, pyminor = sys.version_info[0:2]
    if (pymajor < 3) or (pyminor not in {5, 6, 7, 8}):
        raise RuntimeError(
            '`{}` is unsupported on your version of python, '
            'please use python 3.5 or python 3.6.'.format(__pkg_name__))

    setup(
        name=__pkg_name__,
        version=__version__,
        url='https://github.com/nanoporetech/{}'.format(__pkg_name__),
        author=__author__,
        description=__description__,
        long_description=__long_description__,
        long_description_content_type=__long_description_content_type__,
        python_requires='>=3.5.*,<3.9',
        packages=find_packages(exclude=['*.test', '*.test.*', 'test.*', 'test']),
        package_data={},
        cffi_modules=["build.py:ffibuilder"],
        install_requires=install_requires,
        data_files=[],
        entry_points = {
            'console_scripts': [
                '{0} = {0}.{0}:main'.format(__pkg_name__),
            ]
        },
        scripts=[],
        zip_safe=False,
        cmdclass={},
    )



"""Functionality for aggregating single read methylation calls."""

from collections import Counter, namedtuple
import functools
import multiprocessing
import sys
import threading
import time

import mappy
import numpy as np
from ont_fast5_api.conversion_tools.conversion_utils import get_fast5_file_list
from ont_fast5_api.fast5_interface import get_fast5_file
import pysam

import fast5mod.common
from fast5mod.executor import ProcessPoolExecutor, ThreadPoolExecutor
import libfast5mod

BASECALLANALYSIS = 'Basecall_1D'
MODBASEPATH = 'BaseCalled_template/ModBaseProbs'
FASTQPATH = 'BaseCalled_template/Fastq'
MOTIFS = {
    'dcm': {
        # seq (on fwd strand): [fwd offset, rev offset]
        'CCAGG': ((1, 3), 'MC'),
        'CCTGG': ((1, 3), 'MC')},
    'dam': {
        'GATC': ((1, 2), 'MA')},
    'cpg': {
        'CG': ((0, 1), 'MC')},
    'all': {
        # TODO: these will appear distinct in output, should fix
        'C': ((0, None), 'MC'),
        'G': ((None, 0), 'MC'),
        'A': ((0, None), 'MA'),
        'T': ((None, 0), 'MA')}
    }
Read = namedtuple('Read', ['read_id', 'sequence', 'quality'])


def format_uint8_list(array):
    """Serialize a numpy array to a comma separated list.

    :param array: a uint8 array.

    :returns: a comma separated string.

    """
    ffi = libfast5mod.ffi
    lib = libfast5mod.lib

    pointer = ffi.cast(
        "uint8_t *",
        array.flatten().ctypes.data)
    result = ffi.new("char[{}]".format(4 * len(array)))
    lib.format_uint8_array(pointer, len(array), result)

    string = ffi.string(result).decode()
    return string


def align_read(aligner, read, tags=None):
    """Aligns a `Read` namedtuple to produce a sam record.

    :param aligner: a `mappy.Aligner` instance`
    :param read: a `Read` namedtuple.
    :param tags: list containing additional tags to add to sam record.

    :returns: a (string) sam record.
    """
    try:
        align = list(aligner.map(read.sequence, MD=True, cs=True))[0]
    except IndexError:
        return None
    if align.strand == +1:
        flag = '0'
        seq = read.sequence
        qstring = read.quality
    else:
        flag = '16'
        seq = fast5mod.common.reverse_complement(read.sequence)
        qstring = read.quality[::-1]
    rname = align.ctg
    pos = str(align.r_st + 1)
    mapq = str(align.mapq)
    clip = [
        '' if x == 0 else '{}S'.format(x)
        for x in (align.q_st, len(seq) - align.q_en)]
    if align.strand == -1:
        clip = clip[::-1]
    cigar = clip[0] + align.cigar_str + clip[1]
    NM = 'NM:i:' + str(align.NM)

    # NOTE: tags written without reversal, see below
    if tags is None:
        tags = []
    sam = '\t'.join((
        read.read_id, flag, rname, pos, mapq, cigar,
        '*', '0', '0', seq, qstring, NM, *tags))
    return sam


def unaligned_read(read, tags=None):
    """Create an unaligned sam record for a read.

    :param read: a `Read` namedtuple.
    :param tags: list containing additional tags to add to sam record.

    :returns: a (string) sam record.
    """
    if tags is None:
        tags = []
    sam = '\t'.join((
        read.read_id, '4', '*', '0', '255', '*',
        '*', '0', '0', read.sequence, read.quality, *tags))
    return sam


def hdf_to_sam_worker(fname):
    """Extract and align basecall and methylation data from `.fast5`.

    :param reference: `.fasta` file containing reference sequence(s).
    :param fname: `.'fast5` file containing read data.
    """
    logger = fast5mod.common.get_named_logger('ModExtract')
    logger.info("Processing {}.".format(fname))
    known_mods = {'6mA', '5mC'}
    results = list()
    with get_fast5_file(fname, mode="r") as f5:
        reads = list(f5.get_read_ids())
        logger.debug("Found {} reads for {}.".format(len(reads), fname))
        for read_id in reads:
            read = f5.get_read(read_id)
            latest = read.get_latest_analysis(BASECALLANALYSIS)
            # get modified base data
            mod_base = read.handle[
                'Analyses/{}/{}'.format(latest, MODBASEPATH)]
            mb_names = mod_base.attrs[
                'modified_base_long_names'].decode().split()
            if set(mb_names) == known_mods:
                bases = 'A 6mA C 5mC G T'.split()
            elif '5mC' in mb_names:
                bases = 'A C 5mC G T'.split()
            elif '6mA' in mb_names:
                bases = 'A 6mA C G T'.split()
            else:
                raise IndexError("Unsupported modified base names found.")
            mod_base = mod_base[()].view([(b, 'uint8') for b in bases])
            tags = []
            if '6mA' in mb_names:
                tags.append(
                    'MA:B:C,{}'.format(format_uint8_list(mod_base['6mA'])))
            if '5mC' in mb_names:
                tags.append(
                    'MC:B:C,{}'.format(format_uint8_list(mod_base['5mC'])))
            # get basecalling data
            fastq = read.get_analysis_dataset(latest, FASTQPATH)
            header, sequence, _, qstring = fastq.splitlines()
            # put everything together
            read = Read(read_id, sequence, qstring)
            results.append((read, tags))
    return results


class Extractor(object):
    """Extracts read data from .fast5 files."""

    def __init__(
            self, path, recursive=False, workers=2,
            extractor=hdf_to_sam_worker, max_size=16000):
        """Initialize read data extraction.

        :param path: .fast5 directory path.
        :param recursive: search directory recursively.
        :param workers: number of worker processes.
        :param extractor: callback function to apply to each file.
        :param max_size: approximate maximum size of internal queue.
            When internal queue is at this size. Further read extraction
            will block until results are consumed.

        The results of the extractor can be accessed from the .get() method
        or by iteration.

        """
        self.path = path
        self.recursive = recursive
        self.workers = workers
        self.extractor = extractor
        self.max_size = max_size
        # setting max size for queue has no effect on pausing workers
        # as the queue is filled from a callback in the main thread.
        # Instead we just don't submit new jobs if the queue is above
        # the requested size.
        self.queue = multiprocessing.Queue()
        self.files_processed = 0
        self.total_files = 0

        self.logger = fast5mod.common.get_named_logger("Extractor")
        self.logger.info("Starting worker processes.")
        self._thread = threading.Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()

    def _run(self):
        """Iterate over input files and stores results into internal queue."""
        fast5s = get_fast5_file_list(self.path, recursive=self.recursive)
        self.total_files = len(fast5s)
        self.logger.info("Found {} files to process.".format(self.total_files))
        with ProcessPoolExecutor(
                self.workers, max_workers=self.workers) as executor:
            for fname in fast5s:
                while True:
                    if self.queue.qsize() < self.max_size:
                        future = executor.submit(self.extractor, fname)
                        future.add_done_callback(self._store)
                        break
                    else:
                        time.sleep(1)
        self.queue.put(None)

    def _store(self, future):
        """Store read data from a future."""
        try:
            results = future.result()
            for r in results:
                self.queue.put(r)
        except Exception as e:
            sys.stderr.write(e)
        # https://bugs.python.org/issue27144
        future._result = None
        self.files_processed += 1
        self.logger.info("Extracted {}/{} files.".format(
            self.files_processed, self.total_files))

    def get(self):
        """Get a read from the internal queue."""
        item = self.queue.get()
        if item is not None:
            pass
        else:
            raise StopIteration('All items processed.')
        return item

    def __iter__(self):
        """Iterate over read data from files."""
        while True:
            try:
                yield self.get()
            except StopIteration:
                break


def read_fasta(path):
    """Read a fasta/q file returning tuples as .fast5 extractor.

    :param path: input file
    """
    with pysam.FastxFile(path) as fa:
        for seq in fa:
            seq_len = len(seq.sequence)
            seq = Read('abcdf', seq.sequence, '!' * seq_len)
            scores = np.zeroes(seq_len, dtype='uint8')
            tag = 'MA:B:C,{}'.format(format_uint8_list(scores))
            yield seq, [tag]


def hdf_to_sam(args):
    """Entry point for converting guppy methylcalled fast5s to sam."""
    logger = fast5mod.common.get_named_logger('ModExtract')
    logger.info(
        "NOTE: Mod. base scores are output w.r.t the sequencing direction, "
        "not the aligned read orientation.")

    sys.stdout.write('\t'.join(('@HD', 'VN:1.5', 'SO:unsorted')))
    sys.stdout.write('\n')
    sys.stdout.write('\t'.join((
        '@CO', 'Guppy basecaller mod. base tags are stored w.r.t. the '
        'sequencing direction, they should be reversed for reads '
        'aligning to the reverse strand.\n')))
    if args.reference is None:
        # write unaligned sam
        extractor = Extractor(
            args.path, recursive=args.recursive, workers=args.io_workers)
        for read, tags in extractor:
            sam = unaligned_read(read, tags)
            sys.stdout.write('{}\n'.format(sam))
    else:
        with pysam.FastaFile(args.reference) as fa:
            for name, length in zip(fa.references, fa.lengths):
                sys.stdout.write('@SQ\tSN:{}\tLN:{}\n'.format(name, length))
        sys.stdout.flush()

        # TODO: this will end in a mappy.ThreadBuffer created for each read.
        #       Should grab one per worker thread
        logger.info("Loading alignment index.")
        aligner = functools.partial(
            align_read, mappy.Aligner(args.reference, preset='map-ont'))
        logger.info("Alignment index loaded.")

        with ThreadPoolExecutor(
                max_items=args.workers, max_workers=args.workers) as executor:
            extractor = Extractor(
                args.path, recursive=args.recursive, workers=args.io_workers)
            for read, tags in extractor:
                future = executor.submit(aligner, read, tags)
                future.add_done_callback(_write_sam_future)


def _write_sam_future(future):
    try:
        sam = future.result()
        if sam is not None:
            sys.stdout.write('{}\n'.format(sam))
    except Exception:
        pass
    # https://bugs.python.org/issue27144
    future._result = None


class MotifTracker():
    """Track methylation counts of a motif while parsing a pileup."""

    def __init__(self, ref, region, seq, strand_offsets, tag):
        """Initialize the tracker.

        :param ref: reference sequence.
        :param region: region of reference to parse.
        :param seq: the motif search to examine.
        :param strand_offsets: positions of motif to examine on forward and
            reverse strands (`None` if a particular strand shouldn't be
            examined.)
        :param tag: bam tag to look up data.

        """
        self.motif = seq
        self.ref = ref
        self.region = region
        self.seq = seq
        self.tag = tag
        self.fwd_offset, self.rev_offset = strand_offsets
        self.has_fwd = self.fwd_offset is not None
        self.has_rev = self.rev_offset is not None
        self.loci = self._find_loci(
            self.ref, self.motif, self.region.start, self.region.end)
        self._queue = list()
        self.reset_counters()

    def reset_counters(self):
        """Reset methylation counters to zero."""
        self.meth = Counter()
        self.not_meth = Counter()

    def add(self, is_meth, is_rev=None):
        """Add a methylation/non methylation count to the tallies.

        :param is_meth: boolean, the count to which to add.
        :param is_rev: if given checked against internal state to ensure
            caller is adding to the correct count.

        """
        if is_rev is not None:
            assert is_rev == self.is_rev
        if is_meth:
            self.meth[self.is_rev] += 1
        else:
            self.not_meth[self.is_rev] += 1

    @property
    def summary(self):
        """Return a tuple of counts and meta data.

        The data returned is:

            * reference name
            * reference position
            * motif sequence
            * modified count fwd. strand
            * modified count rev. strand
            * non-modified count fwd. strand
            * non-modified count rev. strand

        """
        return (
            self.region.ref_name, self.index, self.motif,
            self.meth[False], self.meth[True],
            self.not_meth[False], self.not_meth[True])

    def __iter__(self):
        """Return self."""
        return self

    def __next__(self):
        """Iterate over the reference positions which need to be examined."""
        if len(self._queue) == 0:
            self.index = next(self.loci)
            if self.has_fwd:
                self._queue.append((False, self.index + self.fwd_offset))
            if self.has_rev:
                self._queue.append((True, self.index + self.rev_offset))
            if self.has_fwd and self.has_rev:
                if self.rev_offset < self.fwd_offset:
                    self._queue = self._queue[::-1]
        item = self._queue.pop(0)
        self.is_rev, self.pos = item
        self.taken_all = len(self._queue) == 0
        return item

    @staticmethod
    def _find_loci(ref, motif, start, end):
        index = start
        while index < end:
            index = ref.find(motif, index)
            if index == -1:
                break
            yield index
            index += 1


def padded_pileup(pileup, start, end):
    """Pad a pysam pileup with mock object in zero coverage columns.

    :param pileup: pysam pileup.
    :param start: start position.
    :param end: end (exclusive) position.
    """
    class FakeColumn:
        def __init__(self, pos):
            self.pos = pos
            self.pileups = []

    nxt = start
    for col in pileup:
        while nxt != col.pos:
            yield FakeColumn(nxt)
            nxt += 1
        yield col
        nxt += 1

    for x in range(nxt, end):
        yield FakeColumn(x)


def call_methylation(args):
    """Entry point for calling methylation from bam file."""
    logger = fast5mod.common.get_named_logger('Mextract')
    logger.info("Processing {}.".format(args.bam))
    motifs = MOTIFS[args.meth]

    additional_text = ''
    if args.meth == 'all':
        additional_text = (
            " Note motifs are stranded, such that both 'C' and 'G' (and 'A' "
            "and 'T') bases will be seen in the output for 5mC (6mA). "
            "Output with a 'G' ('T' for 6mA) correspond to the reverse "
            "strand. Users should post process the output to join/sum lines "
            "belonging to the same genomic loci.")
    elif args.meth == 'dcm':
        additional_text = (
            " The output is sorted by motif, users may wish to subsequently "
            "sorted by position.")
    logger.info(
        "Calling methylation at {} sites.{}".format(
            args.meth, additional_text))

    references = pysam.FastaFile(args.reference)
    with open(args.output, 'w') as fh:
        with pysam.AlignmentFile(args.bam, "rb") as bam:
            for region in fast5mod.common.get_regions(args.bam, args.regions):
                logger.info("Processing region: {}.".format(region))
                ref = references.fetch(region.ref_name)

                for seq, motif_info in motifs.items():
                    tracker = MotifTracker(ref, region, seq, *motif_info)
                    logger.info("Searching for motif '{}'.".format(seq))
                    try:
                        next(tracker)
                    except StopIteration:
                        continue

                    pileup = bam.pileup(
                        region.ref_name, region.start, region.end)
                    for pileupcolumn in padded_pileup(
                            pileup, region.start, region.end):
                        if (pileupcolumn.pos - region.start) % 1000000 == 0:
                            done = pileupcolumn.pos - region.start
                            pct_done = 100 * done // \
                                (region.end - region.start)
                            logger.info(
                                "Processed {} ref positions ({}%)".format(
                                    done, pct_done))
                        if pileupcolumn.pos > tracker.pos:
                            raise ValueError(
                                "Unexpectedly skipped reference columns.")
                        elif pileupcolumn.pos == tracker.pos:
                            for read in pileupcolumn.pileups:
                                aln = read.alignment
                                if read.is_del \
                                        or read.is_refskip \
                                        or aln.is_reverse != tracker.is_rev:
                                    continue
                                qpos = read.query_position
                                tag = read.alignment.get_tag(tracker.tag)
                                # NOTE: tags are not reversed, see above
                                if not aln.is_reverse:
                                    mscore = tag[qpos]
                                else:
                                    mscore = tag[len(tag) - qpos - 1]
                                if mscore > args.filter[1]:
                                    tracker.add(True)
                                elif mscore < args.filter[0]:
                                    tracker.add(False)
                        else:
                            # skip forward
                            continue

                        if tracker.taken_all:
                            fh.write(
                                '\t'.join(str(x) for x in tracker.summary))
                            fh.write('\n')
                            tracker.reset_counters()
                        try:
                            next(tracker)
                        except StopIteration:
                            break



"""Entry point for fast5mod package."""
import argparse
import logging
import sys

from fast5mod import __version__
import fast5mod.programs


def _log_level():
    """Parser to set logging level and acquire software version/commit."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter, add_help=False)

    modify_log_level = parser.add_mutually_exclusive_group()
    modify_log_level.add_argument(
        '--debug', action='store_const',
        dest='log_level', const=logging.DEBUG, default=logging.INFO,
        help='Verbose logging of debug information.')
    modify_log_level.add_argument(
        '--quiet', action='store_const',
        dest='log_level', const=logging.WARNING, default=logging.INFO,
        help='Minimal logging; warnings only).')

    return parser


def main():
    """Run main entry point."""
    pymajor, pyminor = sys.version_info[0:2]
    if (pymajor < 3) or (pyminor not in {5, 6, 7, 8}):
        raise RuntimeError(
            '`medaka` is unsupported on your version of python, '
            'please use python 3.5 or python 3.6.')

    parser = argparse.ArgumentParser(
        'fast5mod', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    methparser = parser.add_subparsers(
        title='subcommands', description='valid commands',
        help='additional help', dest='command')
    methparser.required = True

    parser.add_argument(
        '--version', action='version',
        version='%(prog)s {}'.format(__version__))

    hdf2samparser = methparser.add_parser(
        'guppy2sam', parents=[_log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help=(
            'Convert Guppy .fast5 files with methylation calls to .sam '
            'file, output to stdout.'))
    hdf2samparser.set_defaults(func=fast5mod.programs.hdf_to_sam)
    hdf2samparser.add_argument(
        'path',
        help='Input path for .fast5 files.')
    hdf2samparser.add_argument(
        '--reference',
        help=(
            '.fasta containing reference sequence(s). If not given output '
            'will be unaligned sam.'))
    hdf2samparser.add_argument(
        '--workers', type=int, default=16,
        help='Number of alignment threads.')
    hdf2samparser.add_argument(
        '--io_workers', type=int, default=4,
        help='Number of .fast5 reader processes.')
    hdf2samparser.add_argument(
        '--recursive', action='store_true',
        help='Search for .fast5s recursively.')

    methcallparser = methparser.add_parser(
        'call', parents=[_log_level()],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help='Call methylation from .bam file.')
    methcallparser.set_defaults(func=fast5mod.programs.call_methylation)
    methcallparser.add_argument(
        'bam',
        help='Input .bam file (via `medaka methylation guppy2sam`).')
    methcallparser.add_argument(
        'reference',
        help='.fasta containing reference sequence(s).')
    methcallparser.add_argument(
        'output',
        help='Output file name.')
    methcallparser.add_argument(
        '--regions', nargs='+',
        help='bam region to process ("chrom:start-end").')
    methcallparser.add_argument(
        '--meth', default='cpg',
        choices=list(fast5mod.programs.MOTIFS.keys()),
        help='methylation type.')
    methcallparser.add_argument(
        '--filter', type=int, nargs=2, default=(64, 128),
        metavar=('upper', 'lower'),
        help=(
            'Upper (lower) score boundary to call canonical (methylated) '
            'base. Scores are in the range [0, 256].'))

    args = parser.parse_args()
    logging.basicConfig(
        format='[%(asctime)s - %(name)s] %(message)s', datefmt='%H:%M:%S',
        level=logging.INFO)
    logger = logging.getLogger(__package__)
    logger.setLevel(args.log_level)

    args.func(args)
